(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('extend'), require('vue')) :
    typeof define === 'function' && define.amd ? define(['extend', 'vue'], factory) :
    (global.KenryForm = factory(global.extend,global.vue));
}(this, (function (extend,vue) { 'use strict';

    extend = extend && extend.hasOwnProperty('default') ? extend['default'] : extend;
    vue = vue && vue.hasOwnProperty('default') ? vue['default'] : vue;

    const emailRegExp = /^[a-z0-9!#$%&'*+\/=?^_`{|}~.-]+@[a-z0-9]([a-z0-9-]*[a-z0-9])?(\.[a-z0-9]([a-z0-9-]*[a-z0-9])?)*$/i; // eslint-disable-line
    const urlRegExp = /^(http\:\/\/|https\:\/\/)(.{4,})$/; // eslint-disable-line

    const email = function (value, attrValue, vnode) {
        return emailRegExp.test(value);
    };
    email._allowNulls = true;

    const number = function (value, attrValue, vnode) {
        return !isNaN(value);
    };
    number._allowNulls = true;

    const url = function (value, attrValue, vnode) {
        return urlRegExp.test(value);
    };
    url._allowNulls = true;

    const validators = {
        email,
        number,
        url,
        required(value, attrValue, vnode) {
            if (attrValue === false) {
                return true;
            }

            if (value === 0) {
                return true;
            }

            if (vnode.data.attrs && typeof vnode.data.attrs.bool !== 'undefined' || vnode.componentOptions && vnode.componentOptions.propsData && typeof vnode.componentOptions.propsData.bool !== 'undefined') {
                // bool attribute is present, allow false pass validation
                if (value === false) {
                    return true;
                }
            }

            if (Array.isArray(value)) {
                return !!value.length;
            }
            return !!value;
        },
        minlength(value, length) {
            return value.length >= length;
        },
        maxlength(value, length) {
            return length >= value.length;
        },
        pattern(value, pattern) {
            const patternRegExp = new RegExp('^' + pattern + '$');
            return patternRegExp.test(value);
        },
        min(value, min, vnode) {
            if ((vnode.data.attrs.type || '').toLowerCase() === 'number') {
                return +value >= +min;
            }
            return value >= min;
        },
        max(value, max, vnode) {
            if ((vnode.data.attrs.type || '').toLowerCase() === 'number') {
                return +max >= +value;
            }
            return max >= value;
        }
    };

    const config = {
        validators,
        isFieldColor: true,
        isAllowMessage: true,
        formComponent: 'kform',
        formTag: 'form',
        Promise: typeof Promise === 'function' ? Promise : null
    };

    const kenryMessages = {
        required: 'This {0} is required.',
        email: 'Please enter a valid email address.',
        url: 'Please enter a valid URL.',
        date: 'Please enter a valid date.',
        dateISO: 'Please enter a valid date (ISO).',
        number: 'Please enter a valid number.',
        digits: 'Please enter only digits.',
        creditcard: 'Please enter a valid credit card number.',
        equalTo: 'Please enter the same value again.',
        accept: 'Please enter a value with a valid extension.',
        maxlength: 'Please enter no more than {0} characters.',
        minlength: 'Please enter at least {0} characters.',
        max: 'Please enter a value less than or equal to {0}.',
        min: 'Please enter a value greater than or equal to {0}.'
    };
    var format = function (str, col) {
        col = typeof col === 'object' ? col : Array.prototype.slice.call(arguments, 1);
        return str.replace(/\{\{|\}\}|\{(\w+)\}/g, function (m, n) {
            if (m === '{{') {
                return '{';
            }
            if (m === '}}') {
                return '}';
            }
            return col[n];
        });
    };
    function activeValidator(fieldState, elm, kMessages, formState) {
        if (!elm) return;
        let allowMessage = !!(formState && formState.isAllowMessage === true);
        let allowColor = !!(formState && formState.isFieldColor === true);
        if (Object.keys(fieldState.$error).length) {
            if (formState.isAllowMessage === undefined) allowMessage = config.isAllowMessage;
            if (formState.isFieldColor === undefined) allowColor = config.isFieldColor;
            if (allowMessage) {
                var nextElement = elm.nextSibling;
                if (nextElement) {
                    var nameNextElm = nextElement.id;
                    if (nameNextElm === 'error_' + elm.name) {
                        nextElement.remove();
                    }
                }
                let condition = Object.keys(fieldState.$error)[0];
                let message = kMessages[condition];
                if (!message) {
                    message = kenryMessages[condition];
                    if (!message) message = '[' + condition + ` message] not exists in messages library, Please check if you have custom message`;
                }
                let name = elm.display || 'field';
                let value = elm.getAttribute(condition);
                if (condition === 'required') value = name;
                if (value === undefined) value = '';
                var errorMessage = format(message, value);
                var elemnt = `<span id="error_${elm.name}" style="color:#ed1c24;margin:5px 0 0 0;display:block;font-size:13px;">${errorMessage}</span>`;
                elm.insertAdjacentHTML('afterEnd', elemnt);
            }
            if (allowColor) {
                elm.setAttribute('style', 'border-color:#de888a !important; background:#fee9ea !important;');
            }
        } else {
            if (allowMessage) {
                var nextElement2 = elm.nextSibling;
                if (nextElement2) {
                    var nameNextElm2 = nextElement2.id;
                    if (nameNextElm2 === 'error_' + elm.name) {
                        nextElement2.remove();
                    }
                }
            }
            if (allowColor) {
                elm.setAttribute('style', 'border-color:none !important; background:none !important;');
            }
        }
    }

    function vModelValue(data) {
        if (data.model) {
            return data.model.value;
        }
        return data.directives.filter(v => v.name === 'model')[0].value;
    }

    function randomId() {
        return Math.random().toString(36).substr(2, 10);
    }

    const kenryFormConfig = `VueFormProviderConfig${randomId()}`;
    const vueFormState = `VueFormProviderState${randomId()}`;
    const vueFormParentForm = `VueFormProviderParentForm${randomId()}`;

    var kform = {
        render(h) {
            if (!this.$attrs.name) {
                console.error('Missing form name');
            } else {
                this.getControls(this.$slots.default);
            }
            return h(this.tag || this.kenryFormConfig.formTag, {
                on: {
                    submit: event => {
                        if (this.state.$pending) {
                            event.preventDefault();
                            this.kenryFormConfig.Promise.all(this.promises).then(() => {
                                this.state._submit();
                                this.$emit('submit', event);
                                this.promises = [];
                            });
                        } else {
                            this.state._submit();
                            this.$emit('submit', event);
                            this.validate();
                        }
                    },
                    reset: event => {
                        this.state._reset();
                        this.$emit('reset', event);
                    }
                },
                // class: this.className,
                attrs: {
                    'novalidate': ''
                }
            }, [this.$slots.default]);
        },
        props: {
            state: {
                type: Object,
                required: true
            },
            tag: String
        },
        inject: { kenryFormConfig },
        provide() {
            return {
                [vueFormState]: this.state,
                [vueFormParentForm]: this
            };
        },
        data: () => ({
            promises: [],
            formMessage: {}
        }),
        created() {
            // set message default
            this.formMessage = kenryMessages;
            let kenryConfig = config;
            if (!this.$attrs.name) {
                console.log('Missing form name');
                return;
            }
            if (!this.state) {
                return;
            }

            // GET CUSTOM VALIDATOR
            if (config.options) {
                let validatorsCs = config.options.validators;
                if (validatorsCs) {
                    let functions = validatorsCs.functions;
                    let cMessage = validatorsCs.messages;
                    if (functions && Object.keys(functions)) {
                        Object.keys(functions).forEach(item => {
                            config.validators[item] = functions[item];
                        });
                    }
                    if (cMessage && Object.keys(cMessage)) {
                        Object.keys(cMessage).forEach(item => {
                            kenryMessages[item] = cMessage[item];
                        });
                    }
                }
                let allowColor = config.options.isFieldColor;
                let allowMessage = config.options.isAllowMessage;
                if (allowColor !== undefined || allowMessage !== null) {
                    config.isFieldColor = allowColor;
                }
                if (allowMessage !== undefined || allowMessage !== null) {
                    config.isAllowMessage = allowMessage;
                }
            }

            // get customer message each page
            if (this.state.messages) {
                this.formMessage = this.state.messages;
            } else {
                // get custome message global
                if (kenryConfig.options && kenryConfig.options.kMessages) {
                    this.formMessage = kenryConfig.options.kMessages;
                }
            }

            const controls = {};
            const state = this.state;
            const formstate = {
                $dirty: false,
                $pristine: true,
                $valid: true,
                $invalid: false,
                $submitted: false,
                $touched: false,
                $untouched: true,
                $focused: false,
                $pending: false,
                $error: {},
                $formName: this.$attrs.name,
                // $controls: [],
                $submittedState: {},
                _id: '',
                _submit: () => {
                    this.state.$submitted = true;
                    this.state._cloneState();
                },
                _cloneState: () => {
                    const cloned = JSON.parse(JSON.stringify(state));
                    delete cloned.$submittedState;
                    Object.keys(cloned).forEach(key => {
                        this.$set(this.state.$submittedState, key, cloned[key]);
                    });
                },
                _addControl: ctrl => {
                    // formstate.$controls.push(ctrl);
                    controls[ctrl.$name] = ctrl;
                    this.$set(state, ctrl.$name, ctrl);
                },
                _removeControl: ctrl => {
                    delete controls[ctrl.$name];
                    this.$delete(this.state, ctrl.$name);
                    this.$delete(this.state.$error, ctrl.$name);
                },
                _validate: () => {
                    Object.keys(controls).forEach(key => {
                        controls[key]._validate(undefined, controls[key], this);
                    });
                },
                _reset: () => {
                    state.$submitted = false;
                    state.$pending = false;
                    state.$submittedState = {};
                    Object.keys(controls).forEach(key => {
                        const control = controls[key];
                        control._hasFocused = false;
                        control._setUntouched();
                        control._setPristine();
                        control.$submitted = false;
                        control.$pending = false;
                    });
                }
            };

            Object.keys(formstate).forEach(key => {
                this.$set(this.state, key, formstate[key]);
            });

            this.$watch('state', () => {
                let isDirty = false;
                let isValid = true;
                let isTouched = false;
                let isFocused = false;
                let isPending = false;
                Object.keys(controls).forEach(key => {
                    const control = controls[key];

                    control.$submitted = state.$submitted;

                    if (control.$dirty) {
                        isDirty = true;
                    }
                    if (control.$touched) {
                        isTouched = true;
                    }
                    if (control.$focused) {
                        isFocused = true;
                    }
                    if (control.$pending) {
                        isPending = true;
                    }
                    if (!control.$valid) {
                        isValid = false;
                        // add control to errors
                        this.$set(state.$error, control.$name, control);
                    } else {
                        this.$delete(state.$error, control.$name);
                    }
                });

                state.$dirty = isDirty;
                state.$pristine = !isDirty;
                state.$touched = isTouched;
                state.$untouched = !isTouched;
                state.$focused = isFocused;
                state.$valid = isValid;
                state.$invalid = !isValid;
                state.$pending = isPending;
            }, {
                deep: true,
                immediate: true
            });
        },
        computed: {},
        methods: {
            reset() {
                this.state._reset();
            },
            validate() {
                this.state._validate();
            },
            getControls(elements) {
                const vm = this;
                if (elements && elements.length) {
                    elements.forEach(item => {
                        if (item.children.length) {
                            let childrens = item.children;
                            childrens.forEach(elm => {
                                if (elm.data) {
                                    if (!elm.data.attrs || !elm.data.attrs.name) {
                                        if (!elm.data.attrs) {
                                            elm.data['attrs'] = {};
                                        }
                                        elm.data.attrs.name = `${elm.tag}_${randomId()}`;
                                        console.log(`tag ${elm.tag} don't have name and it's not validator, Name auto set is ${elm.data.attrs.name}`);
                                    }                                let prevVnode;
                                    let pendingValidators = [];
                                    let fieldstate = {
                                        $name: elm.data.attrs && elm.data.attrs.name ? elm.data.attrs.name : '',
                                        $dirty: false,
                                        $pristine: true,
                                        $valid: true,
                                        $invalid: false,
                                        $touched: false,
                                        $untouched: true,
                                        $focused: false,
                                        $pending: false,
                                        $submitted: false,
                                        $formname: this.$attrs.name,
                                        $error: {},
                                        $attrs: {},
                                        _className: null,
                                        _id: 'kr_' + randomId(),
                                        _setValidatorVadility(validator, isValid) {
                                            if (isValid) {
                                                vm.$delete(this.$error, validator);
                                            } else {
                                                vm.$set(this.$error, validator, true);
                                            }
                                        },
                                        _setValidity(isValid) {
                                            this.$valid = isValid;
                                            this.$invalid = !isValid;
                                        },
                                        _setDirty() {
                                            this.$dirty = true;
                                            this.$pristine = false;
                                        },
                                        _setPristine() {
                                            this.$dirty = false;
                                            this.$pristine = true;
                                        },
                                        _setTouched() {
                                            this.$touched = true;
                                            this.$untouched = false;
                                        },
                                        _setUntouched() {
                                            this.$touched = false;
                                            this.$untouched = true;
                                        },
                                        _setFocused(value) {
                                            this.$focused = typeof value === 'boolean' ? value : false;
                                            if (this.$focused) {
                                                this._setHasFocused();
                                            } else {
                                                this._setTouched();
                                            }
                                        },
                                        _setHasFocused() {
                                            this._hasFocused = true;
                                        },
                                        _hasFocused: false,
                                        _validators: {},
                                        _validate(vnode, fieldstate, thisform) {
                                            if (!vnode) {
                                                vnode = prevVnode;
                                            } else {
                                                prevVnode = vnode;
                                            }
                                            fieldstate.$pending = true;
                                            let isValid = true;
                                            const value = vModelValue(vnode.data);

                                            const pending = {
                                                promises: [],
                                                names: []
                                            };

                                            pendingValidators.push(pending);

                                            let attrs = vnode.data.attrs || {};
                                            const childvm = vnode.componentInstance;
                                            if (childvm && childvm._vfValidationData_) {
                                                attrs = extend({}, attrs, childvm._vfValidationData_);
                                            }

                                            const propsData = vnode.componentOptions && vnode.componentOptions.propsData ? vnode.componentOptions.propsData : {};

                                            Object.keys(fieldstate._validators).forEach(validator => {
                                                // when value is empty and current validator is not the required validator, the field is valid
                                                if ((value === '' || value === undefined || value === null) && validator !== 'required') {
                                                    fieldstate._setValidatorVadility(validator, true);
                                                    // return early, required validator will
                                                    // fall through if it is present
                                                    return;
                                                }

                                                const attrValue = typeof attrs[validator] !== 'undefined' ? attrs[validator] : propsData[validator];
                                                const isFunction = typeof fieldstate._validators[validator] === 'function';

                                                // match vue behaviour, ignore if attribute is null or undefined. But for type=email|url|number and custom validators, the value will be null, so allow with _allowNulls
                                                if (isFunction && (attrValue === null || typeof attrValue === 'undefined') && !fieldstate._validators[validator]._allowNulls) {
                                                    return;
                                                }

                                                if (attrValue) {
                                                    fieldstate.$attrs[validator] = attrValue;
                                                }

                                                const result = isFunction ? fieldstate._validators[validator](value, attrValue, vnode) : vm.custom[validator];

                                                if (typeof result === 'boolean') {
                                                    if (result) {
                                                        fieldstate._setValidatorVadility(validator, true);
                                                    } else {
                                                        isValid = false;
                                                        fieldstate._setValidatorVadility(validator, false);
                                                    }
                                                    if (thisform && thisform.state.$submitted) activeValidator(fieldstate, vnode.elm, thisform.formMessage, thisform.state);
                                                } else {
                                                    pending.promises.push(result);
                                                    pending.names.push(validator);
                                                    vm.vueFormParentForm.promises.push(result);
                                                }
                                            });

                                            if (pending.promises.length) {
                                                vm.kenryFormConfig.Promise.all(pending.promises).then(results => {
                                                    // only concerned with the last promise results, in case
                                                    // async responses return out of order
                                                    if (pending !== pendingValidators[pendingValidators.length - 1]) {
                                                        // console.log('ignoring old promise', pending.promises);
                                                        return;
                                                    }

                                                    pendingValidators = [];

                                                    results.forEach((result, i) => {
                                                        const name = pending.names[i];
                                                        if (result) {
                                                            fieldstate._setValidatorVadility(name, true);
                                                        } else {
                                                            isValid = false;
                                                            fieldstate._setValidatorVadility(name, false);
                                                        }
                                                    });
                                                    fieldstate._setValidity(isValid);
                                                    fieldstate.$pending = false;
                                                });
                                            } else {
                                                fieldstate._setValidity(isValid);
                                                fieldstate.$pending = false;
                                            }
                                        }
                                    };
                                    if (elm.data.directives) {
                                        if (elm.data.directives[0] && elm.data.directives[0]['name'] === 'model') {
                                            elm.data.directives.push({ name: 'kenry-form-validator', value: { fieldstate: fieldstate, config: this.state, validators: config.validators } });
                                            elm.data.attrs['kenry-form-validator'] = '';
                                        } else {
                                            console.error('missing v-model');
                                        }
                                    }
                                }
                            });
                        }
                    });
                }
            }
        }
    };

    const debouncedValidators = {};

    function addValidators(attrs, validators$$1, fieldValidators) {
        Object.keys(attrs).forEach(attr => {
            const prop = attr === 'type' ? attrs[attr].toLowerCase() : attr.toLowerCase();

            if (validators$$1[prop] && !fieldValidators[prop]) {
                fieldValidators[prop] = validators$$1[prop];
            }
        });
    }

    function compareChanges(vnode, oldvnode, validators$$1) {
        let hasChanged = false;
        const attrs = vnode.data.attrs || {};
        const oldAttrs = oldvnode.data.attrs || {};
        const out = {};

        if (vModelValue(vnode.data) !== vModelValue(oldvnode.data)) {
            out.vModel = true;
            hasChanged = true;
        }

        Object.keys(validators$$1).forEach(validator => {
            if (attrs[validator] !== oldAttrs[validator]) {
                out[validator] = true;
                hasChanged = true;
            }
        });

        // if is a component
        if (vnode.componentOptions && vnode.componentOptions.propsData) {
            const attrs = vnode.componentOptions.propsData;
            const oldAttrs = oldvnode.componentOptions.propsData;
            Object.keys(validators$$1).forEach(validator => {
                if (attrs[validator] !== oldAttrs[validator]) {
                    out[validator] = true;
                    hasChanged = true;
                }
            });
        }

        if (hasChanged) {
            return out;
        }
    }

    var kenryFormValidator = {
        name: 'kenry-form-validator',
        bind(el, binding, vnode) {
            const { fieldstate } = binding.value;
            const { validators: validators$$1 } = binding.value;
            el.formName = fieldstate.$formname;
            const attrs = vnode.data.attrs || {};
            // const inputName = getName(vnode);
            // add validators
            addValidators(attrs, validators$$1, fieldstate._validators);

            fieldstate._validate(vnode, fieldstate);
            // activeValidator(fieldstate, vnode);
            // native listeners
            el.addEventListener('blur', () => {
                fieldstate._setFocused(false);
            }, false);
            el.addEventListener('focus', () => {
                fieldstate._setFocused(true);
            }, false);

            // component listeners
            const vm = vnode.componentInstance;
            if (vm) {
                vm.$on('blur', () => {
                    fieldstate._setFocused(false);
                });
                vm.$on('focus', () => {
                    fieldstate._setFocused(true);
                });

                vm.$once('vf:addFocusListeners', () => {
                    el.addEventListener('focusout', () => {
                        fieldstate._setFocused(false);
                    }, false);
                    el.addEventListener('focusin', () => {
                        fieldstate._setFocused(true);
                    }, false);
                });
            }
            binding.value.config._addControl(fieldstate);
        },

        update(el, binding, vnode, oldVNode) {
            if (!el.formName) {
                console.log('Missing form name');
                return;
            }
            if (!el.name) {
                console.log('Missing element name');
                return false;
            }

            const formStates = vnode.context[el.formName];
            const fieldstate = formStates[el.name];
            const { validators: validators$$1 } = binding.oldValue;
            const changes = compareChanges(vnode, oldVNode, validators$$1);

            let attrs = vnode.data.attrs || {};
            const vm = vnode.componentInstance;
            if (vm && vm._vfValidationData_) {
                attrs = extend({}, attrs, vm[vm._vfValidationData_]);
            }

            if (!changes) {
                return;
            }
            if (changes.vModel) {
                // re-validate all
                if (attrs.debounce) {
                    fieldstate.$pending = true;
                    debouncedValidators[fieldstate._id](fieldstate, vnode);
                } else {
                    if (fieldstate._hasFocused) {
                        fieldstate._setDirty();
                    }
                    fieldstate._validate(vnode, fieldstate);
                    let systemMessage = kenryMessages;
                    debugger;
                    if (formStates.messages) {
                        systemMessage = formStates.messages;
                    } else {
                        if (config.options && config.options.kMessages) {
                            systemMessage = config.options.kMessages;
                        }
                    }
                    activeValidator(fieldstate, el, systemMessage, formStates);
                }
            }
            // else {
            //     fieldstate._validate(vnode);
            // }
        }
    };

    function KenryFormBase(options) {
        const c = extend(true, {}, config, options);
        this.provide = () => ({
            [kenryFormConfig]: c
        });
        this.components = {
            [c.formComponent]: kform
        };
        this.directives = { kenryFormValidator };
    }

    class KenryForm extends KenryFormBase {
        constructor() {
            super();
        }
        static install(Vue, options) {
            config.options = options;
            Vue.mixin(new this(options));
        }
        static get installed() {
            return !!this.install.done;
        }
        static set installed(val) {
            this.install.done = val;
        }
    }

    KenryFormBase.call(KenryForm);
    // temp fix for vue 2.3.0
    KenryForm.options = new KenryForm();

    return KenryForm;

})));
